import requests
import time
import os
import re

def sizeof_fmt(num, suffix='B'):
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

def req_file_size(req):
    try:
        return int(req.headers['content-length'])
    except:
        return 0

def makeSafeFilename(inputFilename):
    # Set here the valid chars
    ret = ''
    map = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ._'
    for char in inputFilename:
        if char in map:
           ret += char
        else:
           if char==' ':
                ret+=''
           else:
                ret += ''
    return ret

def get_url_file_name(url,req):
    try:
        if "Content-Disposition" in req.headers.keys():
            try:
                name = str(re.findall("filename=(.+)", req.headers["Content-Disposition"])[0])
                return name
            except:
                name = str(req.headers["Content-Disposition"]).replace('attachment; ','')
                name = name.replace('filename=','').replace('"','')
                return name
        else:
            import urllib
            urlfix = urllib.parse.unquote(url,encoding='utf-8', errors='replace')
            tokens = str(urlfix).split('/');
            return tokens[len(tokens)-1]
    except:
        import urllib
        urlfix = urllib.parse.unquote(url,encoding='utf-8', errors='replace')
        tokens = str(urlfix).split('/');
        return tokens[len(tokens)-1]
    return ''

def get_file_size(file):
    file_size = os.stat(file)
    return file_size.st_size

def createID(count=8):
    from random import randrange
    map = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    id = ''
    i = 0
    while i<count:
        rnd = randrange(len(map))
        id+=map[rnd]
        i+=1
    return id


def nice_time(delta):
    """
    Gets time delta in seconds and returns a pretty string
    representing it in the format of 1w2d9h16m44s
    """

    weeks = 0
    days = 0
    hours = 0
    minutes = 0
    seconds = 0

    seconds_in_minute = 60
    seconds_in_hour = 60 * seconds_in_minute
    seconds_in_day = 24 * seconds_in_hour
    seconds_in_week = 7 * seconds_in_day

    weeks = delta / seconds_in_week
    if weeks != 0:
        delta -= weeks * seconds_in_week

    days = delta / seconds_in_day
    if days != 0:
        delta -= days * seconds_in_day

    hours = delta / seconds_in_hour
    if hours != 0:
        delta -= hours * seconds_in_hour

    minutes = delta / seconds_in_minute
    if minutes != 0:
        delta -= minutes * seconds_in_minute

    seconds = delta

    out = ""
    if seconds:
        out = "%ss" % seconds + out
    if minutes:
        out = "%sm" % minutes + out
    if hours:
        out = "%sh" % hours + out
    if days:
        out = "%sd" % days + out
    if weeks:
        out = "%sw" % weeks + out

    if out == "":
        return "just now"
    return out


import threading
from .utils import createID

class StoppableThread(threading.Thread):
    def __init__(self,  *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

class ObigramThread(object):
    def __init__(self,targetfunc=None,args=(),update=None):
        self.id = createID(12)
        self.handle = threading.Thread(target=targetfunc, args=args)
        self.update = update
        self.tstore = {}
        pass
    def start(self):
        self.handle.start()
    def stop(self):
        self.handle.join()
        pass
    def store(self,name,obj):
        self.tstore[name] = obj
    def getStore(self,name):
        try:
            return self.tstore[name]
        except:pass
        return None


import os
from io import BufferedReader
import time

import requests
from .utils import get_file_size,get_url_file_name,req_file_size

class MonitorReader(object):
    def __init__(self,total=100):
        self.chunk_por = 0
        self.chunkrandom = 100
        self.total = total
        self.time_start = time.time()
        self.time_total = 0
        self.size_per_second = 0
        self.clock_start = time.time()
    def can_step(self,size=0):
        self.chunk_por += size
        self.size_per_second += size
        tcurrent = time.time() - self.time_start
        self.time_total += tcurrent
        self.time_start = time.time()
        if self.time_total>=1:
            self.clock_time = (self.total - self.chunk_por) / (self.size_per_second)
            return True
        return False
    def cleanup(self):
        self.time_total = 0
        self.size_per_second = 0

class FileProgressReader(BufferedReader):
    def __init__(self,filepath='',read_size=1024,progress_func=None,progress_args=None,self_in = None,normalize=False):
        self.file_path = filepath
        self.file_name = str(self.file_path).split('/')[-1]
        self.file = open(filepath,'rb')
        self.read_size = read_size
        self.progress_func = progress_func
        self.progress_args = progress_args
        self.file_size = get_file_size(self.file_path)
        self.chunk_por = 0
        self.chunkrandom = 100
        self.time_start = time.time()
        self.time_total = 0
        self.size_per_second = 0
        self.clock_start = time.time()
        self.self_in = self_in
        self.normalize = normalize
        BufferedReader.__init__(self,self.file,self.read_size)

    def __len__(self):
        return self.file_size

    def read(self, size=1024):
        if self.normalize:
            chunk = self.file.read(size)
        else:
            chunk = self.file.read(self.read_size)
        try:
            self.chunk_por += len(chunk)
            self.size_per_second+=len(chunk)
            tcurrent = time.time() - self.time_start
            self.time_total += tcurrent
            self.time_start = time.time()
            if self.time_total>=1:
               self.clock_time = (self.file_size - self.chunk_por) / (self.size_per_second)
               if self.progress_func:
                  self.progress_func(self.self_in,self.file_name,self.chunk_por,self.file_size,self.size_per_second,self.clock_time,self.progress_args)
               self.time_total = 0
               self.size_per_second = 0
        except:pass
        return chunk

    def flush(self,data=None):
        self.file.flush()

    def close(self):
        try:
            if self.progress_func:
                  self.progress_func(self.self_in,self.file_name,self.file_size,self.file_size,self.self.file_size,self.clock_time,self.progress_args)
        except:pass
        self.file.close()

class FileUrlProgressReader(BufferedReader):
    def __init__(self,url='',read_size=1024,progress_func=None,progress_args=None,self_in = None):
        self.url = url
        self.chunk_por = 0
        self.req = self.make_get()
        self.file_name = get_url_file_name(self.url,self.req)
        self.read_size = read_size
        self.progress_func = progress_func
        self.progress_args = progress_args
        self.file_size = req_file_size(self.req)
        self.chunkrandom = 100
        self.time_start = time.time()
        self.time_total = 0
        self.size_per_second = 0
        self.clock_start = time.time()
        self.self_in = self_in
        temp = open('temp.url','w')
        temp.write(f'{self.url}')
        temp.close()
        self.file = open('temp.url','rb')
        BufferedReader.__init__(self,self.file,self.read_size)

    def make_get(self,read=False):
        resp = None
        if not read:
            resp = requests.get(self.url,stream = True,allow_redirects=True)
        else:
            headers = {}
            headers['Range'] = f'bytes={self.chunk_por}-{self.read_size}'
            resp = requests.get(self.url,stream = True,allow_redirects=True,headers=headers)
        return resp

    def __len__(self):
        return self.file_size

    def read(self, size=1024):
        chunk = b''
        resp = self.make_get(True)
        chunk += resp.content
        self.chunk_por += len(chunk)
        self.size_per_second+=len(chunk)
        tcurrent = time.time() - self.time_start
        self.time_total += tcurrent
        self.time_start = time.time()
        if self.time_total>=1:
           self.clock_time = (self.file_size - self.chunk_por) / (self.size_per_second)
           if self.progress_func:
              self.progress_func(self.self_in,self.file_name,self.chunk_por,self.file_size,self.size_per_second,self.clock_time,self.progress_args)
           self.time_total = 0
           self.size_per_second = 0
        return chunk

    def flush(self,data=None):
        self.file.flush()

    def close(self):
        try:
            if self.progress_func:
                  self.progress_func(self.self_in,self.file_name,self.file_size,self.file_size,self.self.file_size,self.clock_time,self.progress_args)
            os.unlink('temp.url')
        except:pass
        self.file.close()

#file = FileUrlProgressReader(url='https://www.7-zip.org/a/7z2201-x64.exe')
#chunk = file.read()
#print('')



import os
import codecs

from pyobigram.inline import inlineKeyboardMarkupArray,inlineKeyboardButton

class TTUI(object):
    def __init__(self, path='tuis/'):
        self.path = path
        if not os.path.exists(self.path):
            os.mkdir(self.path)

    def parse_markups(self,markups):
        result = []
        for markup in markups:
            newarray = []
            for item in markup:
                if item['type'] == 'button':
                   text = ''
                   url = ''
                   callback = ''
                   if 'text' in item:
                       text = item['text']
                   if 'url' in item:
                       url = item['url']
                   if 'callback' in item:
                       callback = item['callback']
                   button = inlineKeyboardButton(text,url,callback)
                   newarray.append(button)
            result.append(newarray)
        if len(result)>0:
           return inlineKeyboardMarkupArray(result)
        else:
            return None


    def _exec_lines(self,lines,args={},sections={}):
        result = ''
        recolect = False
        recolect_lines = []
        calle = ''
        for line in lines:
                if '##' in line:continue
                if line=='':continue
                try:
                    if '@if' in line and not recolect or '@for' in line and not recolect:
                        calle = line
                        recolect = True
                        continue
                    if recolect:
                        if '@endif' in line and '@if' in calle:
                            recolect = False
                            if calle!='':
                                calle = str(calle).replace('@if ','').replace('\r','').replace('\n','')
                                if 'not ' in calle:
                                    calle = str(calle).replace('not ','')
                                    if not args[calle]:
                                       result += self._exec_lines(recolect_lines,args)
                                elif args[calle]:
                                    result += self._exec_lines(recolect_lines,args)
                            recolect_lines.clear()
                            continue
                        elif '@endfor' in line and '@for' in calle:
                            recolect = False
                            if calle!='':
                                calle = str(calle).replace('@for ','').replace('\r','').replace('\n','')
                                tokens = str(calle).split(' in ')
                                key = tokens[0]
                                iter = tokens[1]
                                args[key] = None
                                args['for_index'] = 0
                                if iter in args:
                                   for item in args[iter]:
                                       try:
                                           for itemin in item:
                                               args[f'{key}.{itemin}'] = item[itemin]
                                           pass
                                       except:pass
                                       args[key] = item
                                       result += self._exec_lines(recolect_lines,args)
                                       args['for_index'] += 1
                                args.pop(key)
                                args.pop(for_index)
                            recolect_lines.clear()
                            continue
                        recolect_lines.append(line)
                        continue
                    if '@>>' in line:
                        keys = []
                        tokens = str(line).split(' ')
                        for t in tokens:
                            if '@>>' in t:
                                tremp = str(t).split('@>>')[1]
                                keys.append(str(tremp).replace('\r','').replace('\n','').replace('@','').replace('>>',''))
                        line_result = line
                        for key in keys:
                            if key in args:
                               line_result = str(line_result).replace(f'@>>{key}',str(args[key]))
                            else:
                               line_result = str(line_result).replace(f'@>>{key}','none')
                        result += line_result + '\n'
                        continue
                    if '@onejump' in line:
                        result = str(result).replace('\n\n','\n')
                        continue
                    if '@jmpto_sect' in line:
                        section = str(line).replace('@jmpto_sect ','').replace('\n','').replace('\r','')
                        if section in sections:
                           result += self._exec_lines(sections[section],args=args,sections=sections)
                        continue
                    result += line + '\n'
                except Exception as ex:
                    pass
        return result

    def _parse_markup(self,line):
        line = str(line).replace('\r','').replace('\n','')
        markup = {}
        if '@' in line:
            line = str(line).replace('@','')
            tokens = str(line).split(' ')
            type = tokens[0]
            markup['type'] = type
            i = 1
            value = ''
            recolect = False
            while i < len(tokens)+1:
                try:
                    item = tokens[i]
                except:
                    item = '='
                    recolect = True
                if '=' in item:
                    if recolect:
                       recolect = False
                       tvalues = str(value).replace("'",'').split('=')
                       key = tvalues[0]
                       val = tvalues[1]
                       markup[key] = val
                       value = ''
                    else:
                        recolect = True
                    value += item
                    i+=1
                    continue
                if recolect:
                    value += item + ' '
                    i+=1
                    continue
                i+=1
            pass
        return markup

    def _get_sections_markups(self,lines):
        sections = {}
        recolect = False
        recolect_lines = []
        markups = []
        calle = ''
        for line in lines:
                if '##' in line:continue
                if line=='':continue
                try:
                    if '@section' in line and not recolect or '@markup' in line and not recolect:
                        calle = line
                        recolect = True
                        continue
                    if recolect:
                        if '@endsection' in line and '@section' in calle:
                            recolect = False
                            if calle!='':
                               calle = str(calle).replace('@section ','').replace('\r','').replace('\n','')
                               sections[calle] =  []
                               for re in recolect_lines:
                                   sections[calle].append(re)
                            recolect_lines.clear()
                            continue
                        if '@endmarkup' in line and '@markup' in calle:
                            recolect = False
                            addmarkups = []
                            if len(recolect_lines)>0:
                               for re in recolect_lines:
                                   if '@jmplist' in re:
                                       markups.append(addmarkups)
                                       addmarkups = []
                                       continue
                                   addmarkups.append(self._parse_markup(re))
                                   pass
                               markups.append(addmarkups)
                            recolect_lines.clear()
                            continue
                        recolect_lines.append(line)
                except:pass
        return sections,markups

    def render(self,name='',args={},section='__base'):
        path = self.path
        result = None
        define = ''
        markups = []
        if name!='':
            tmp = codecs.open(f'{path}{name}.tui','r','utf8')
            result = tmp.read()
            tmp.close()
        if result:
            for arg in args:
                result = str(result).replace(f'@>>{arg}',str(args[arg]))

            lines = str(result).split('\n')
            sections,markups = self._get_sections_markups(lines)
        
            if section in sections:
                lines = sections[section]
            else:
                lines = []

            if len(lines)>0:
                line0 = lines[0]
                if '@define ' in line0:
                    lines.pop(0)
                    define = str(line0).replace('@define ','').replace('\n','').replace('\r','')
            result = self._exec_lines(lines,args=args,sections=sections)
        if len(markups)>0:
           return define,result,markups
        else:
           return define,result


#Inline Buttons
def inlineKeyboardMarkup(**params):
    buttons = []
    for item in params:
        buttons.append(params[item])
    return {'inline_keyboard':buttons}

def inlineKeyboardMarkupArray(paramms):
    return {'inline_keyboard':paramms}

def inlineKeyboardButton(text='text',url='',callback_data=''):
    result = {'text':text}
    if url!='':
       result['url'] = url
    if callback_data!='':
       result['callback_data'] = callback_data
    return result

#Inline Queries
def inlineQueryResultArticle(id=0,title='',
                             text='',
                             url='',hide_url=False,
                             thumb_url='',
                             thumb_width=10,thumb_height=10,
                             parse_mode=''):
    result = {'type':'article',
            'id':id,
            'title':title,
            'input_message_content':{'message_text':text},
            'url':url,
            'hide_url':hide_url,
            'thumb_url':thumb_url,
            'thumb_width':thumb_width,
            'thumb_height':thumb_height}
    if parse_mode!='':
        result['parse_mode'] = parse_mode
    return result

def inlineQueryResultDocument(id=0,title='',
                              caption='',text='',description='',
                              url='',hide_url=False,
                              thumb_url='',
                              thumb_width=10,thumb_height=10,
                              parse_mode=''):
    result = {'type':'article',
            'id':id,
            'title':title,
            'input_message_content':{'message_text':text},
            'document_url':url,
            'hide_url':hide_url,
            'thumb_url':thumb_url,
            'thumb_width':thumb_width,
            'thumb_height':thumb_height}
    if parse_mode!='':
        result['parse_mode'] = parse_mode
    return result


import requests
import json
import threading
try:
    from types import SimpleNamespace as Namespace
except ImportError:
    from argparse import Namespace
import time
import codecs
import uuid
import os

from .threads import ObigramThread
from .utils import get_url_file_name,req_file_size,createID,makeSafeFilename
from .readers import FileProgressReader,FileUrlProgressReader

from typing import Dict, cast
from telethon.tl.custom import Message
from telethon.tl.types import TypeInputPeer, InputPeerChannel, InputPeerChat, InputPeerUser

from telethon import TelegramClient
from telethon.tl.types import InputDocument
from telethon import functions, types
from .paralleltransfer import download_file,create_stream

import asyncio

class ObigramClient(object):
    def __init__(self,token,api_id='',api_hash=''):
        self.token = token
        self.path = 'https://api.telegram.org/bot' + token + '/'
        self.files_path = 'https://api.telegram.org/file/bot' + token + '/'
        self.runing = False
        self.funcs = {}
        self.update_id = 0
        self.onmessage = None
        self.oninline = None
        self.SendFileTypes = {'document':'SendDocument','video':'SendVideo'}
        self.this_thread = None
        self.threads = {}
        # api telegram mtproto
        self.api_id = api_id
        self.api_hash = api_hash
        self.mtproto = None
        self.transfer = None
        self.loop = None
        self.Way = False
        self.store = {}
        self.callback_funcs = {}
        self.temps = []

    def newMTP(self,name='obigram'):
        if self.api_id!='' and self.api_hash!='':
            try:
                os.unlink(f'{name}.session')
                os.unlink(f'{name}.session-journal')
            except:pass
            self.mtproto = TelegramClient(name,api_id=self.api_id,api_hash=self.api_hash)
            self.mtproto.start(bot_token=self.token)
            try:
                    self.loop = asyncio.get_runing_loop();
            except:
                    try:
                        self.loop = asyncio.get_event_loop();
                    except:self.loop = self.mtproto.loop
        
    def startNewThread(self,targetfunc=None,args=(),update=None):
        self.this_thread = ObigramThread(targetfunc=targetfunc,args=args,update=update)
        self.threads[self.this_thread.id] = self.this_thread
        self.this_thread.start()
        pass

    def parseUpdate(self,update):
        parse = str(update).replace('sender_chat','chat')
        parse = str(parse).replace('from','sender')
        parse = str(parse).replace('my_chat_member','message')
        parse = str(parse).replace('document','file')
        parse = str(parse).replace('video','file')
        parse = str(parse).replace('photo','file')
        return parse

    def way(self,way=False):self.Way=way

    def run_forever_loop(self,args=None):
        #self.loop.run_forever()
        pass

    def run(self):
        self.newMTP()
        self.runing = True
        if self.loop:
            self.startNewThread(self.run_forever_loop)
        while self.runing:
            while self.Way==True:pass
            try:
                getUpdateUrl = self.path + 'getUpdates?offset=' + str(self.update_id+1)
                update = requests.get(getUpdateUrl)
                update = self.parseUpdate(str(update.text))
                updates = json.loads(update, object_hook = lambda d : Namespace(**d)).result

                if len(updates) > 0:
                    self.update_id = updates[-1].update_id

                try:
                    for func in self.funcs:
                        for update in updates:
                                if func in update.message.text:
                                    self.startNewThread(self.funcs[func],(update,self),update)
                except:pass

                try:
                        for update in updates:
                            try:
                                if update.inline_query:
                                    if self.oninline:
                                        self.startNewThread(self.oninline,(update,self),update)
                                    break
                            except:
                                try:
                                    if update.callback_query:
                                        for callback in self.callback_funcs:
                                            if callback in update.callback_query.data:
                                                update.callback_query.data = str(update.callback_query.data).replace(callback,'')
                                                self.startNewThread(self.callback_funcs[callback],
                                                                    (update.callback_query, self),
                                                                    update.callback_query)
                                                break
                                except:
                                    if self.onmessage:
                                        self.startNewThread(self.onmessage,(update,self),update)
                except Exception as ex:print(str(ex))

            except Exception as ex:
                print(str(ex))
            pass
        self.threads.clear()
        pass

    def send_message(self,chat_id=0,text='',parse_mode='',reply_markup=None,reply_to_message_id=None):
        try:
            text=text.replace('%', '%25')
            text=text.replace('#', '%23')
            text=text.replace('+', '%2B')
            text=text.replace('*', '%2A')
            text=text.replace('&', '%26')
            sendMessageUrl = self.path + 'sendMessage?chat_id=' + str(chat_id) + '&text=' + text + '&parse_mode=' + parse_mode
            payload = {'reply_markup': reply_markup}
            jsonData = {}
            if reply_to_message_id:
                payload['reply_to_message_id'] = reply_to_message_id
                jsonData['reply_to_message_id'] = reply_to_message_id
            if reply_markup:
                jsonData = payload
            result = requests.get(sendMessageUrl,json=jsonData).text
            result = self.parseUpdate(result)
            return json.loads(result, object_hook = lambda d : Namespace(**d)).result
        except Exception as ex:
            print(str(ex))
            pass
        return None

    def delete_message(self,message):
        try:
            deleteMessageUrl = self.path + 'deleteMessage?chat_id='+str(message.chat.id)+'&message_id='+str(message.message_id)
            result = requests.get(deleteMessageUrl).text
            result = self.parseUpdate(result)
            return json.loads(result, object_hook = lambda d : Namespace(**d)).result
        except:pass
        return None

    def edit_message(self,message,text='',parse_mode='',reply_markup=None):
        if message:
            try:
                text=text.replace('%', '%25')
                text=text.replace('#', '%23')
                text=text.replace('+', '%2B')
                text=text.replace('*', '%2A')
                text=text.replace('&', '%26')
                editMessageUrl = self.path+'editMessageText?chat_id='+str(message.chat.id)+'&message_id='+str(message.message_id)+'&text=' + text + '&parse_mode=' + parse_mode
                payload = {'reply_markup': reply_markup}
                jsonData = {}
                if reply_markup:
                    jsonData = payload
                result = requests.get(editMessageUrl,json=jsonData).text
                result = self.parseUpdate(result)
                parse = json.loads(result, object_hook = lambda d : Namespace(**d))
                sussesfull = False
                try: 
                    sussesfull = parse.ok and parse.result 
                    if sussesfull == False:
                         print('Warning EditMessage: '+str(parse.description))
                except: pass
                message.text = text
                return message
            except: pass
        return None


    def send_file(self,chat_id,file,type='document',caption=None,reply_to_message_id=None,reply_markup=None,thumb=None):
        sendDocumentUrl = self.path + self.SendFileTypes[type]
        of = codecs.open(file)
        if thumb:
            thumbfile = codecs.open(thumb)
        payload_files = {type:(file,of)}
        payload_data = {'chat_id':chat_id}
        if thumb:
            payload_files['thumb'] = (thumb,thumbfile)
        if reply_to_message_id:
            payload_data['reply_to_message_id'] = reply_to_message_id
        if reply_markup:
            payload_data['reply_markup'] = reply_markup
        if caption:
            payload_data['caption'] = caption
        result = requests.post(sendDocumentUrl,files=payload_files,data=payload_data).text
        result = self.parseUpdate(result)
        of.close()
        if thumb:
            thumbfile.close()
        parse = json.loads(result, object_hook = lambda d : Namespace(**d))
        return parse.result

    def get_file(self,file_id):
        getFileUrl = self.path + 'getFile?file_id=' + str(file_id)
        result = requests.get(getFileUrl).text
        result = self.parseUpdate(result)
        parse = json.loads(result, object_hook = lambda d : Namespace(**d)).result
        return parse

    def download_file(self,file_id=0,destname='',progressfunc=None,args=None):
        reqFile = self.getFile(file_id)
        downloadUrl = self.files_path + str(reqFile.file_path)
        req = requests.get(downloadUrl, stream = True,allow_redirects=True)
        if req.status_code == 200:
            file_wr = open(destname,'wb')
            chunk_por = 0
            chunkrandom = 100
            total = reqFile.file_size
            time_start = time.time()
            time_total = 0
            size_per_second = 0
            for chunk in req.iter_content(chunk_size = 1024):
                    chunk_por += len(chunk)
                    size_per_second+=len(chunk);
                    tcurrent = time.time() - time_start
                    time_total += tcurrent
                    time_start = time.time()
                    file_wr.write(chunk)
                    if time_total>=1:
                        if progressfunc:
                            progressfunc(destname,chunk_por,total,size_per_second,args)
                        time_total = 0
                        size_per_second = 0
            file_wr.close()
        return destname

    def mtp_download_file(self,message,dest_path='',progress_func=None,progress_args=None):
        async def asyncexec_download():
            peer = InputPeerUser(user_id=message.chat.id, access_hash=0)
            forward = cast(Message, await self.mtproto.get_messages(entity=peer, ids=message.message_id))
            #forward = await self.mtproto.forward_messages(message.sender.id,message.message_id,from_peer=message.sender.id)
            #await forward.delete()
            filename = forward.file.id + forward.file.ext
            fsize = forward.file.size
            if forward.file.name:
                filename = forward.file.name
            filename = makeSafeFilename(filename)
            output = dest_path
            if '/' in output:
                output += '/'
            output += filename
            await download_file(self.mtproto,forward.media,output,fsize,progress_func,progress_args,self)
            self.store[message.message_id] = output
        self.loop.run_until_complete(asyncexec_download())
        output = None
        while not output:
            try:
                output = self.store[message.message_id]
                self.store.pop(message.message_id)
            except:pass
            pass
        return output

    async def async_get_file_info(self,message):
        peer = InputPeerUser(user_id=message.chat.id, access_hash=0)
        forward = cast(Message, await self.mtproto.get_messages(entity=peer, ids=message.message_id))
        # forward = await self.mtproto.forward_messages(message.sender.id,message.message_id,from_peer=message.sender.id)
        # await forward.delete()
        filename = forward.file.id + forward.file.ext
        fsize = forward.file.size
        if forward.file.name:
            filename = forward.file.name
        filename = makeSafeFilename(filename)
        return {'fname':filename,'fsize':fsize,'location':message}

    def mtp_get_file_info(self,message):
        async def asyncexec_download():
            peer = InputPeerUser(user_id=message.chat.id, access_hash=0)
            forward = cast(Message, await self.mtproto.get_messages(entity=peer, ids=message.message_id))
            #forward = await self.mtproto.forward_messages(message.sender.id,message.message_id,from_peer=message.sender.id)
            #await forward.delete()
            filename = forward.file.id + forward.file.ext
            fsize = forward.file.size
            if forward.file.name:
                filename = forward.file.name
            filename = makeSafeFilename(filename)
            self.store[message.message_id] = {'fname':filename,'fsize':fsize,'location':message}
        self.loop.run_until_complete(asyncexec_download())
        output = None
        while not output:
            try:
                output = self.store[message.message_id]
                self.store.pop(message.message_id)
            except:pass
            pass
        return output

    def mtp_gen_message(self,chat_id,msg_id):
        class Chat(object):
            def __init__(self, id):
                self.id = id
                pass
        class Message(object):
            def __init__(self,chat_id,msg_id):
                self.chat = Chat(chat_id)
                self.message_id = msg_id
                pass
        return Message(chat_id,msg_id)

    async def async_get_info_stream(self,message):
        peer = InputPeerUser(user_id=message.chat.id, access_hash=0)
        forward = cast(Message, await self.mtproto.get_messages(entity=peer, ids=message.message_id))
        # forward = await self.mtproto.forward_messages(message.sender.id,message.message_id,from_peer=message.sender.id)
        # await forward.delete()
        filename = forward.file.id + forward.file.ext
        fsize = forward.file.size
        if forward.file.name:
            filename = forward.file.name
        filename = makeSafeFilename(filename)
        body = create_stream(self.mtproto, forward.media, fsize)
        return {'fname':filename,'body':body,'fsize':fsize}

    def mtp_gen_stream(self,message):
        async def asyncexec_download():
            peer = InputPeerUser(user_id=message.chat.id, access_hash=0)
            forward = cast(Message, await self.mtproto.get_messages(entity=peer, ids=message.message_id))
            #forward = await self.mtproto.forward_messages(message.sender.id,message.message_id,from_peer=message.sender.id)
            #await forward.delete()
            filename = forward.file.id + forward.file.ext
            fsize = forward.file.size
            if forward.file.name:
                filename = forward.file.name
            filename = makeSafeFilename(filename)
            body = create_stream(self.mtproto,forward.media,fsize)
            self.store[message.message_id] = {'fname':filename,'body':body,'fsize':fsize}
        self.loop.run_until_complete(asyncexec_download())
        output = None
        while not output:
            try:
                output = self.store[message.message_id]
                self.store.pop(message.message_id)
            except:pass
            pass
        return output

    def mtp_send_file(self,sender,filepath,progress_func=None,progress_args=None):
        upload_id = createID()
        async def asyncexec_upload():
            file = FileProgressReader(filepath,progress_func=progress_func,progress_args=progress_args,self_in=self,normalize=True)
            send = await self.mtproto.send_file(sender.id,file)
            self.store[upload_id] = send
            pass
        self.loop.run_until_complete(asyncexec_upload())
        output = None
        while not output:
            try:
                output = self.store[upload_id]
            except:pass
            pass
        return output

    def mtp_send_file_from_url(self,sender,url,progress_func=None,progress_args=None):
        upload_id = createID()
        async def asyncexec_upload():
            file = FileUrlProgressReader(url=url,progress_func=progress_func,progress_args=progress_args,self_in=self)
            send = await self.mtproto.send_file(sender.id,file)
            self.store[upload_id] = send
            pass
        self.loop.run_until_complete(asyncexec_upload())
        output = None
        while not output:
            try:
                output = self.store[upload_id]
            except:pass
            pass
        return output

    def forward_message(self,sender_id,message):
        async def asyncexec_forward():
            forward = await self.mtproto.forward_messages(sender_id,message.message_id,from_peer=message.sender.id)
            self.store[sender_id] = forward
            pass
        self.loop.run_until_complete(asyncexec_forward())
        forward = None
        while not forward:
            try:
                forward = self.store[sender_id]
                self.store.pop(sender_id)
            except:pass
            pass
        return forward

    def answer_inline(self,inline_query_id=0,result=[]):
        answerUrl = self.path + 'answerInlineQuery'
        payload = { 'inline_query_id' : inline_query_id,'results':result}
        result = requests.post(answerUrl,json=payload).text
        result = self.parseUpdate(result)
        parse = json.loads(result, object_hook = lambda d : Namespace(**d))
        sussesfull = False
        try: 
            sussesfull = parse.ok and parse.result 
            if sussesfull == False:
                 print('Error InlineAnswer: '+str(parse.description))
        except: pass
        return sussesfull

    def contain_file(self,message):
        try:
            if message.file:
                return True
        except:pass
        return False

    def on (self,cmd,func):self.funcs[cmd] = func
    def onMessage (self,func):self.onmessage = func
    def onInline(self,func):self.oninline = func
    def onCallbackData(self,callback_data,func):self.callback_funcs[callback_data] = func
